from pathlib import Path
from typing import Union
from typing import List
import multiprocessing as mp

import numpy as np
from tqdm import tqdm
import decord

from .util import FR_Module
from .helpers import VideoReaderWrapper, BufferedVideoReader

## Define Dataset class as a subclass of utils.FR_Module
class Dataset_videos(FR_Module):
    """
    Class for handling video data.
    Imports videos using decord and generates lazy video reader
     objects for each video.
    RH 2022
    """
    def __init__(
        self,
        bufferedVideoReader: BufferedVideoReader=None,
        paths_videos: Union[str, List[str]]=None,
        contiguous: bool=False,
        frame_rate_clamp: float=None,
        verbose: Union[bool, int]=1,
    ):
        """
        Initialize the Dataset_videos object.

        Args:
            bufferedVideoReader (BufferedVideoReader, optional):
                Optional. If None, then paths_videos must be specified.
                BufferedVideoReader object to use for reading videos.
                Object should have all the desired videos loaded.
            paths_videos (str or list of str):
                Optional. If None, then bufferedVideoReader must be
                 specified.
                Path(s) to video(s) to load.
            contiguous (bool):
                Whether the videos should be treated as contiguous.
                I.e. Whether the first frame of the next video should be
                 considered to be the subsequent frame of the last video.
            frame_rate_clamp (float):
                If None then the frame rate metadata from the videos will
                 be used.
                If a float then the frame rate will be set to this value.
            verbose (bool or int):
                Whether to print progress messages.
                0: No messages
                1: Warnings
                2: Info
        """
        ## Imports
        super().__init__()

        ## Set variables
        self.contiguous = bool(contiguous)
        self.verbose = int(verbose)

        ## Determine whether to use bufferedVideoReader or paths_videos
        self._videoDataType = 'BufferedVideoReader' if bufferedVideoReader is not None else 'paths_videos'
        
        ### Assert that either bufferedVideoReader or paths_videos is specified
        assert bufferedVideoReader is not None or paths_videos is not None, "FR ERROR: bufferedVideoReader or paths_videos must be specified"
        
        ## Workflow if method is 'paths_videos'
        ### Assert that if using 'paths_videos', it is either a list of strings or a string
        if self._videoDataType == 'paths_videos':
            if isinstance(paths_videos, list): 
                assert all([isinstance(path, str) for path in paths_videos]), "FR ERROR: paths_videos must be a string or list of strings to paths of videos"  
            else:
                assert isinstance(paths_videos, str), "FR ERROR: paths_videos must be a string or list of strings to paths of videos"
            ## If paths_videos is a string, convert it to a list of strings
            self.paths_videos = paths_videos if isinstance(paths_videos, list) else [paths_videos]

            ## Assert that all paths_videos exist
            exists_paths_videos = [bool(Path(path).exists()) for path in self.paths_videos]
            assert all(exists_paths_videos), f"FR ERROR: paths_videos must exist. The following paths do not exist: {[path for path, exists in zip(self.paths_videos, exists_paths_videos) if not exists]}"


            ## Load videos
            print("FR: Loading lazy video reader objects...") if self.verbose > 1 else None
            decord.bridge.set_bridge('torch')
            self.videos = [VideoReaderWrapper(path_video, ctx=decord.cpu(0), num_threads=mp.cpu_count()) for path_video in tqdm(self.paths_videos, disable=(self.verbose < 2))]

            ## make video metadata dataframe
            print("FR: Collecting video metadata...") if self.verbose > 1 else None
            self.metadata = {"paths_videos": self.paths_videos}
            self.num_frames, self.frame_rate, self.frame_height_width, self.num_channels = [], [], [], []
            for v in tqdm(self.videos):
                self.num_frames.append(int(len(v)))
                self.frame_rate.append(float(v.get_avg_fps()))
                frame_tmp = v[0]
                self.frame_height_width.append([int(n) for n in frame_tmp.shape[:2]])
                self.num_channels.append(int(frame_tmp.shape[2]))
            self.metadata["num_frames"] = self.num_frames
            self.metadata["frame_rate"] = self.frame_rate
            self.metadata["frame_height_width"] = self.frame_height_width
            self.metadata["num_channels"] = self.num_channels
                

            ## Assert that all videos must have at least one frame
            assert all([n > 0 for n in self.metadata["num_frames"]]), "FR ERROR: All videos must have at least one frame"
            ## Assert that all videos must have the same shape
            assert all([n == self.metadata["frame_height_width"][0] for n in self.metadata["frame_height_width"]]), "FR ERROR: All videos must have the same shape"
            ## Assert that all videos must have the same number of channels
            assert all([n == self.metadata["num_channels"][0] for n in self.metadata["num_channels"]]), "FR ERROR: All videos must have the same number of channels"

        ## Workflow if method is 'BufferedVideoReader'
        elif self._videoDataType == 'BufferedVideoReader':
            ## Assert that bufferedVideoReader is a BufferedVideoReader object
            print('printing this line helps the bufferedVideoReader object load properly.  ', type(bufferedVideoReader), type(BufferedVideoReader), BufferedVideoReader.__class__, isinstance(bufferedVideoReader, BufferedVideoReader)) if self.verbose > 1 else None   ## line needed sometimes for next assert to work
            assert isinstance(bufferedVideoReader, BufferedVideoReader), "FR ERROR: bufferedVideoReader must be a BufferedVideoReader object"
            ## Set self.videos to bufferedVideoReader
            self.videos = bufferedVideoReader.video_readers
            ## Set self.paths_videos to bufferedVideoReader.paths_videos
            self.paths_videos = bufferedVideoReader.paths_videos
            ## Set self.metadata to bufferedVideoReader.metadata
            self.metadata = bufferedVideoReader.metadata

        ## set frame rate
        if frame_rate_clamp is None:
            frame_rates = self.metadata["frame_rate"]
            ## warn if any video's frame rate is very different from others
            max_diff = float((np.max(frame_rates) - np.min(frame_rates)) / np.mean(frame_rates))
            print(f"FR WARNING: max frame rate difference is large: {max_diff*100:.2f}%") if ((max_diff > 0.1) and (self.verbose > 0)) else None
            self.frame_rate = float(np.median(frame_rates))
        else:
            self.frame_rate = float(frame_rate_clamp)

        self.num_frames_total = int(np.sum(self.metadata["num_frames"]))
        self.frame_height_width = self.metadata["frame_height_width"][0]
        self.num_channels = self.metadata["num_channels"][0]
        self.paths_videos = [str(path) for path in self.paths_videos]  ## ensure paths are strings

        self.example_image = self.videos[0][0]

        ## For FR_Module compatibility
        self.config = {
            "paths_videos": self.paths_videos,
            "contiguous": contiguous,
            "frame_rate_clamp": frame_rate_clamp,
            "verbose": verbose,
        }
        self.run_info = {
            "frame_rate": self.frame_rate,
            "num_frames_total": self.num_frames_total,
            "frame_height_width": self.frame_height_width,
            "num_channels": self.num_channels,
            "metadata": self.metadata,
        }
        self.run_data = {
            "example_image": self.example_image,
        }
        ## Append the self.run_info data to self.run_data
        # self.run_data.update(self.run_info)

    def __repr__(self):
        return f"Dataset_videos, num_videos={len(self.paths_videos)}, num_frames_total={self.num_frames_total}, frame_rate={self.frame_rate}, frame_height_width={self.frame_height_width}, num_channels={self.num_channels}"

    ## Define methods for loading and handling videos
    def __getitem__(self, index): return self.videos[index]
    def __len__(self): return len(self.videos)
    def __iter__(self): return iter(self.videos)
    def __next__(self): return next(self.videos)



from typing import Union
from pathlib import Path

import numpy as np
import torch
import hdfdict
from tqdm import tqdm

from .util import FR_Module
from . import helpers

class VQT_Analyzer(FR_Module):
    """
    A class for generating normalized spectrograms for point
     displacement traces. The spectrograms are generated using
     the Variable Q-Transform (VQT) algorithm.
    RH 2022
    """
    def __init__(
        self,
        params_VQT: dict={
            'Fs_sample': 90,
            'Q_lowF': 3,
            'Q_highF': 20,
            'F_min': 1,
            'F_max': 40,
            'n_freq_bins': 55,
            'win_size': 501,
            'plot_pref': False,
            'downsample_factor': 4,
            'padding': 'valid',
            'DEVICE_compute': 'cpu',
            'DEVICE_return': 'cpu',
            'return_complex': False,
            'filters': None, 
            'plot_pref': False,
            'progressBar': True,
        },
        normalization_factor: float=0.99,
        spectrogram_exponent: float=1.0,
        one_over_f_exponent: float=1.0,
        verbose: int=1,
    ):
        super().__init__()

        ## Set attributes
        self._normalization_factor = float(normalization_factor)
        self._spectrogram_exponent = float(spectrogram_exponent)
        self._one_over_f_exponent = float(one_over_f_exponent)
        self._verbose = int(verbose)
        self.spectrograms = None
        self.x_axis = None
        self.freqs = None

        self._demo_spectrogram = None

        ## Initalize VQT filters
        self.VQT = helpers.VQT(**params_VQT)

        ## For FR_Module compatibility
        self.config = {
            'params_VQT': params_VQT,
            'normalization_factor': normalization_factor,
            'spectrogram_exponent': spectrogram_exponent,
            'verbose': verbose,
        }
        self.run_info = {
        }
        self.run_data = {
            'VQT': {key: self.VQT.__dict__[key] for key in ['filters', 'wins']},
            'frequencies': self.VQT.freqs,
        }
        ## Append the self.run_info data to self.run_data
        # self.run_data.update(self.run_info)
        # self.run_data['config'] = self.config

    def cleanup(self):
        """
        Delete the large data objects that are no longer needed.
        """
        import gc
        print(f"FR: Deleting all attributes")
        while len(self.__dict__.keys()) > 0:
            key = list(self.__dict__.keys())[0]
            del self.__dict__[key]
        gc.collect()
        gc.collect()


    def transform(self, points_tracked: np.ndarray, point_positions: np.ndarray):
        """
        Transform a single batch of points into spectrograms.

        Args:
            points_tracked (np.ndarray):
                A 3D numpy array of shape(n_frames, n_points, 2)
                 containing the tracked points.
            point_positions (np.ndarray):
                A 2D numpy array of shape(n_points, 2) containing the positions

        Returns:
            spectrograms (np.ndarray):
                A 4D numpy array of shape(n_points, n_freq_bins, n_frames, 2)
        """
        ## Prepare traces
        point_positions = self._prepare_pointPositions(point_positions)
        ## Compute spectrograms
        spec, x_axis, freqs = self.VQT(self._prepare_displacements(points_tracked, point_positions))
        spec_rs = self._normalize_spectrogram(spec)
        return spec_rs.cpu().numpy(), x_axis.cpu().numpy(), freqs


    def transform_all(self, points_tracked: dict, point_positions: np.ndarray):
        """
        Generate spectrograms from the tracked points.
        Reshape the inputs, run the VQT, and normalize the spectrograms.

        Args:
            points_tracked (dict):
                A dictionary of 3D numpy arrays of shape(n_frames, n_points, 2)
                 containing the tracked points.
            point_positions (np.ndarray):
                A 2D numpy array of shape(n_points, 2) containing the positions
                 of the tracked points.

        Set Attributes:
            self.spectrograms (dict):
                A list of spectrograms for each element in idx_points.
            self.run_data (dict):
                Updates with self.spectrograms and self.point_positions.
        """
        ## Check inputs
        self._check_inputs(points_tracked, point_positions)

        ## Prepare traces
        print(f"Preparing traces. Reshaping and subtracting offsets...") if self._verbose > 1 else None
        self.point_positions = self._prepare_pointPositions(point_positions)
        ## Compute spectrograms
        print(f"Computing spectrograms...") if self._verbose > 1 else None
        self.spectrograms, self.x_axis, self.freqs = {}, {}, {}
        for key, points in tqdm(points_tracked.items(), disable=not self._verbose > 1, desc='Computing spectrograms', leave=True, position=1):
            self.spectrograms[key], self.x_axis[key], self.freqs[key] = self.transform(points, point_positions)

        ## Update self.run_data
        self.run_data.update({
            'spectrograms': self.spectrograms,
            'x_axis': self.x_axis,
            'point_positions': self.point_positions,
        })
        
    def _check_inputs(self, points_tracked: dict, point_positions: np.ndarray):
        """
        Check the inputs to the transform function.

        Args:
            points_tracked (dict):
                A dictionary of 3D numpy arrays of shape(n_frames, n_points, 2)
                 containing the tracked points.
            point_positions (np.ndarray):
                A 2D numpy array of shape(n_points, 2) containing the positions
                 of the tracked points.
        """
        ## Assertions
        ### Assert that the points_tracked dict is valid
        assert isinstance(points_tracked, (dict, hdfdict.hdfdict.LazyHdfDict)), f"points_tracked must be a dict or hdfdict.hdfdict.LazyHdfDict, not {type(points_tracked)}. See docstring for details."
        ### Assert that points_tracked contains 3D numpy arrays of shape(n_frames, n_points, 2)
        for key, value in points_tracked.items():
            assert isinstance(value, np.ndarray), f"points_tracked must contain numpy arrays, not {type(value)}. See docstring for details."
            assert value.ndim == 3, f"points_tracked must contain 3D numpy arrays, not {value.ndim}D. Shape should be (n_frames, n_points, 2). See docstring for details."
            assert value.shape[2] == 2, f"points_tracked must contain 3D numpy arrays of shape(n_frames, n_points, 2), not {value.shape}. See docstring for details."
        ### Assert that point_positions is a 2D numpy array of shape(n_points, 2)
        assert isinstance(point_positions, np.ndarray), f"point_positions must be a numpy array, not {type(point_positions)}. See docstring for details."
        assert point_positions.ndim == 2, f"point_positions must be a 2D numpy array, not {point_positions.ndim}D. Shape should be (n_points, 2). See docstring for details."
        assert point_positions.shape[1] == 2, f"point_positions must be a 2D numpy array of shape(n_points, 2), not {point_positions.shape}. See docstring for details."

    ## Prepare normalization function
    def _normalize_spectrogram(self, spectrogram):
        """
        Normalize spectrogram by dividing by the total power
            across all frequencies at each time point.

        Args:
            spectrogram (torch.Tensor):
                A single spectrogram
                Spectrogram should be of shape: (n_points, n_freq_bins, n_frames)  

        Returns:
            spectrogram (tuple or torch.Tensor):
                A single normalized spectrogram of same shape as input.
        """
        ## Check inputs
        s = spectrogram

        if torch.is_complex(s) == False:
            s_exp = s ** self._spectrogram_exponent
            s_mean = torch.mean(torch.sum(s_exp , dim=1) , dim=0)  ## Mean of the summed power across all frequencies and points. Shape (n_frames,)
            s_norm =  s_exp / ((self._normalization_factor * s_mean[None,None,:]) + (1-self._normalization_factor))  ## Normalize the spectrogram by the mean power across all frequencies and points. Shape (n_points, n_freq_bins, n_frames)
        
        elif torch.is_complex(s) == True:
            s_mag = torch.abs(s)
            s_phase = torch.angle(s)
            s_exp = s_mag ** self._spectrogram_exponent
            s_mean = torch.mean(torch.sum(s_exp , dim=1) , dim=0)  ## Mean of the summed power across all frequencies and points. Shape (n_frames,)
            s_mag_norm = s_exp / ((self._normalization_factor * s_mean[None,None,:]) + (1-self._normalization_factor))  ## Normalize the spectrogram by the mean power across all frequencies and points. Shape (n_points, n_freq_bins, n_frames)
            s_norm = torch.polar(s_mag_norm, s_phase)
        
        ## Do 1/f correction
        s_norm = s_norm * torch.as_tensor(self.VQT.freqs[None,:,None] ** self._one_over_f_exponent, dtype=torch.float32) if self._one_over_f_exponent != 0 else s_norm
        s_norm_rs = s_norm.reshape(2, int(s_norm.shape[0]/2), s_norm.shape[1], s_norm.shape[2])
        return s_norm_rs

    def _prepare_displacements(self, traces, point_positions):
        """
        A function to reshape and subtract offsets from traces.
        Reshape from (n_frames, n_points, 2) to (n_points * 2, n_frames)
        Subtract the point_positions from the traces.

        Args:
            traces (np.ndarray):
                A 3D numpy array of shape(n_frames, n_points, 2) containing the
                 tracked points.
            point_positions (np.ndarray):
                A 2D numpy array of shape(n_points * 2) containing the positions
                 of the tracked points.

        Returns:
            displacements (np.ndarray):
                A 2D numpy array of shape(n_points * 2, n_frames) containing the
                 displacements of the tracked points.
        """
        out = torch.as_tensor(traces.reshape((traces.shape[0], -1), order='F').T, dtype=torch.float32) - point_positions[:,None]
        return out

    def _prepare_pointPositions(self, point_positions):
        """
        A function to reshape point_positions from (n_points, 2) to (n_points * 2)
        """
        return torch.as_tensor(point_positions.reshape((-1,), order='F').T, dtype=torch.float32)


    def demo_transform(
        self, 
        points_tracked: dict, 
        point_positions: np.ndarray, 
        idx_point: list=[0,], 
        name_points: str='0', 
        plot: bool=True,
    ):
        """
        Do a test transform to check that the spectrograms are being
         computed correctly and look good.
        Use the plot to check it out visually.

        Args:
            points_tracked (dict):
                A dictionary of 3D numpy arrays of shape(n_frames, n_points, 2)
                 containing the tracked points.
            point_positions (np.ndarray):
                A 2D numpy array of shape(n_points, 2) containing the positions
                 of the tracked points.
            name_pointsTracked (str):
                A string to use to index the points_tracked dict.
            idx_point (list):
                Index of point to transform and return or plot.
            plot (bool):
                Whether to plot the spectrograms.

        Returns:
            spectrograms (dict):
                A list of spectrograms for each element in idx_points.
        """
        ## Check inputs
        self._check_inputs(points_tracked, point_positions)

        ## Transform
        spec, x_axis, freqs = self.transform(points_tracked[name_points][:,idx_point,:][:,None,:], point_positions[idx_point,:][None,...])
        spec = spec.squeeze(1)
        # spec, x_axis, freqs = out['spec'][:,0,:,:], out['x_axis'], out['freqs']
        print(f"Demo spectrogram shape: {spec.shape}")

        ## Plot
        if plot:
            import matplotlib.pyplot as plt
            x_0 = x_axis[0] / self.VQT.Fs_sample
            x_N = x_axis[-1] / self.VQT.Fs_sample
            fig, axs = plt.subplots(2, 1, figsize=(10, 5))
            axs[0].imshow(np.abs(spec[0,:,:]), aspect='auto', origin='lower', cmap='hot')
            axs[1].imshow(np.abs(spec[1,:,:]), aspect='auto', origin='lower', cmap='hot')
            axs[0].set_title(f'Spectrogram of x and y displacements of point {idx_point}')
            axs[1].set_xlabel('Time (s)')
            axs[0].set_ylabel('Frequency')
            ## set yticks values
            yticks = np.linspace(0, len(freqs), num=8, endpoint=False, dtype=int)
            yticklabels = np.round(freqs[yticks], 2)
            axs[0].set_yticks(yticks)
            axs[0].set_yticklabels(yticklabels)
            axs[1].set_yticks(yticks)
            axs[1].set_yticklabels(yticklabels)

            plt.show()


        ## Compute size of all output spectrograms
        ### Get the output shape for each spectrogram
        def _get_avgpool1d_downsample_length(length, kernel_size, stride, padding):
            return int((length + 2 * padding - kernel_size) / stride + 1)
        shapes_pt = [p.shape for p in points_tracked.values()]
        shapes_spec = [(s[2], len(self.VQT.freqs), s[1], _get_avgpool1d_downsample_length(s[0], kernel_size=int(self.VQT.downsample_factor), stride=self.VQT.downsample_factor, padding=0)) for s in shapes_pt]
        sizes_spec = [np.prod(s) * spec.itemsize / 1e9 for s in shapes_spec]
        print(f'Total size of all spectrograms: {np.sum(sizes_spec):.8f} GB') if self._verbose > 1 else None
        print(f'Individual spectrogram sizes (in GB): {sizes_spec}') if self._verbose > 1 else None            

        return spec, x_axis, freqs

    
    def __repr__(self): return f"{self.__class__.__name__}( normalization_factor={self._normalization_factor}, spectrogram_exponent={self._spectrogram_exponent}, VQT={self.VQT}, verbose={self._verbose} )"
    def __getitem__(self, index): return self.spectrograms(index)
    def __len__(self): return len(self.spectrograms)
    def __iter__(self): return iter(self.spectrograms.items())    
    def __call__(self, points_tracked: dict, point_positions: np.ndarray, name_points: str='0'): 
        """
        Compute spectrograms for each point in points_tracked.
        Wrapper for self.transform()
        See self.transform() for details.
        """
        return self.transform(points_tracked, point_positions, name_points)


from typing import Union
import time

import numpy as np
from tqdm import tqdm
import decord
import cv2
import torch
import scipy.sparse

from .util import FR_Module
from .rois import ROIs
from .helpers import BufferedVideoReader
from .visualization import FrameVisualizer

## Define class for performing point tracking using optical flow
class PointTracker(FR_Module):
    def __init__(
        self,
        buffered_video_reader: BufferedVideoReader,
        point_positions: np.ndarray,
        rois_masks: ROIs=None,
        contiguous: bool=False,
        params_optical_flow: dict={
                        "method": "lucas_kanade", ## method for optical flow. Only "lucas_kanade" is supported for now.
                        "mesh_rigidity": 0.005,  ## Rigidity of mesh. Changes depending on point spacing.
                        "relaxation": 0.5,  ## How quickly points relax back to their original position.
                        "kwargs_method": {
                            "winSize": (15,15),  ## Size of window to use for optical flow
                            "maxLevel": 2,  ## Maximum number of pyramid levels
                            "criteria": (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 0.03),  ## Stopping criteria for optical flow optimization
                        },
                    },
        params_outlier_handling: dict={
                        'threshold_displacement': 25,  ## Maximum displacement between frames, in pixels.
                        'framesHalted_before': 30,  ## Number of frames to halt tracking before a violation.
                        'framesHalted_after': 30,  ## Number of frames to halt tracking after a violation.
                    },
        visualize_video: bool=False,
        params_visualization: dict={
                        'alpha':1.0,
                        'point_sizes':1,
        },
        verbose: Union[bool, int]=1,
    ):
        """
        Prepare for point tracking.
        Sets up parameters for optical flow, makes initial points,
         to track, and prepares masks for the video.

        Args:
            buffered_video_reader (BuffereVideoReader):
                A BufferedVideoReader object, containing the videos to track.
                Object comes from fr.helpers.BufferedVideoReader
            point_positions (2D array of floats):
                An array of points to track.
                Each row is a point, and the columns are the x and y coordinates.
                Can be made using the fr.rois.ROIs class, and the 
                 ROIs.point_positions attribute.
            rois_masks (2D array of booleans or list of 2D arrays of booleans):
                An ROI is a 2D array of booleans, where True indicates a pixel
                 that is within the region of interest.
                If a list of ROIs is provided, then each ROI will be tracked
                 separately.
            contiguous (bool, optional):
                Whether or not videos should be treated as contiguous.
                If True, the first frame of each video will be treated
                 as the next frame of the previous video.
                If False, the point tracking will be restarted for each
                 video.
            params_optical_flow (dict, optional):
                Parameters for optical flow.
                If None, the following parameters will be used:
                    params_optical_flow = {
                        "method": "lucas_kanade",  ## method for optical flow. Only "lucas_kanade" is supported for now.
                        "point_spacing": 10,  ## spacing between points, in pixels
                        "mesh_rigidity": 0.005,  ## Rigidity of mesh. Changes depending on point spacing.
                        "relaxation": 0.5,  ## How quickly points relax back to their original position.
                        "kwargs_method": {
                            "winSize": (15,15),  ## Size of window to use for optical flow
                            "maxLevel": 2,  ## Maximum number of pyramid levels
                            "criteria": (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 0.03),  ## Stopping criteria for optical flow optimization
                        },
                    }
                See https://docs.opencv.org/3.4/d4/dee/tutorial_optical_flow.html
                 and https://docs.opencv.org/3.4/dc/d6b/group__video__track.html#ga473e4b886d0bcc6b65831eb88ed93323
                 for more information about the lucas kanade optical flow parameters.
            params_outlier_handling (dict, optional):
                Parameters for outlier handling.
                Outliers/violations are frames when a point goes beyond a
                 threshold distance from its original position. For these
                 frames, the violating point has its velocity frozen for
                 'framesHalted_before' frames before and 'framesHalted_after'
                 frames after the violation.
                If None, the following parameters will be used:
                    params_outlier_handling = {
                        'threshold_displacement': 25,  ## Maximum displacement between frames, in pixels.
                        'framesHalted_before': 30,  ## Number of frames to halt tracking before a violation.
                        'framesHalted_after': 30,  ## Number of frames to halt tracking after a violation.
                    }
            visualize_video (bool, optional):
                Whether or not to visualize the video.
                If on a server or system without a display, this should be False.
            params_visualization (dict, optional):
                Parameters for visualization.
                If None, the following parameters will be used:
                    params_visualization = {
                        'alpha':1.0,
                        'point_sizes':1,
                        'writer_cv2':None,
                    }
                See fr.visualization.FrameVisualizer for more information.
                Leave out 'points_colors' as this is reserved for outlier coloring.
            verbose (bool or int, optional):
                Whether or not to print progress updates.
                0: no progress updates
                1: warnings
                2: all info
        """
        ## Imports
        super().__init__()
        
        ## Set variables
        self._contiguous = bool(contiguous)
        self._verbose = int(verbose)
        self._visualize_video = bool(visualize_video)
        self._params_visualization = params_visualization.copy()
        self._params_outlier_handling = params_outlier_handling.copy()

        ## Assert that buffered_video_reader is a fr.helpers.BufferedVideoReader object
        type(buffered_video_reader), isinstance(buffered_video_reader, BufferedVideoReader)  ## line needed sometimes for next assert to work
        assert isinstance(buffered_video_reader, BufferedVideoReader), "buffered_video_reader must be a fr.helpers.BufferedVideoReader object."
        ## Assert that point_positions is a 2D array of floats
        assert isinstance(point_positions, np.ndarray), "point_positions must be a 2D array of floats. Use the fr.rois.ROIs class to generate using .point_positions attribute."
        ## Assert that the rois variables are either 2D arrays or lists of 2D arrays
        if isinstance(rois_masks, np.ndarray):
            rois_masks = [rois_masks]
        ## Assert that params_optical_flow is a dict
        assert isinstance(params_optical_flow, dict), "FR ERROR: params_optical_flow must be a dict"
        ## Assert that params_outlier_handling is a dict
        assert isinstance(params_outlier_handling, dict), "FR ERROR: params_outlier_handling must be a dict"
        ## Assert that params_visualization is a dict
        assert isinstance(params_visualization, dict), "FR ERROR: params_visualization must be a dict"
        
        ## Define default parameters
        params_optFlow_default = {
                "method": "lucas_kanade",
                "mesh_rigidity": 0.005,
                "mesh_n_neighbors": 10,
                "relaxation": 0.5,
                "kwargs_method": {
                    "winSize": [15,15],
                    "maxLevel": 2,
                    "criteria": [cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 0.03],
                },
            }
        params_outlierHandling_default = {
                'threshold_displacement': 25,
                'framesHalted_before': 30,
                'framesHalted_after': 30,
            }
        params_visualization_default = {
                'alpha':1.0,
                'point_sizes':1,
            }

        ## Fill in default parameters
        ## if the parameter was passed as an arg then place it in the args dict, otherwise use the default
        ### params_optical_flow
        params_missing = {key: params_optFlow_default[key] for key in params_optFlow_default if key not in params_optical_flow}
        print(f"FR WARNING: Following parameters for optical flow were not specified and will be set to default values: {params_missing}") if ((self._verbose > 0) and (len(params_missing) > 0)) else None
        self.params_optical_flow = {**params_optical_flow, **params_missing}
        ### params_outlier_handling
        params_missing = {key: params_outlierHandling_default[key] for key in params_outlierHandling_default if key not in params_outlier_handling}
        print(f"FR WARNING: Following parameters for outlier handling were not specified and will be set to default values: {params_missing}") if ((self._verbose > 0) and (len(params_missing) > 0)) else None
        self.params_outlier_handling = {**params_outlier_handling, **params_missing}
        ### params_visualization
        params_missing = {key: params_visualization_default[key] for key in params_visualization_default if key not in params_visualization}
        print(f"FR WARNING: Following parameters for visualization were not specified and will be set to default values: {params_missing}") if ((self._verbose > 0) and (len(params_missing) > 0)) else None
        self.params_visualization = {**params_visualization, **params_missing}

        ## Retrive point positions
        self.point_positions = point_positions
        self.num_points = self.point_positions.shape[0]
        print(f"FR: {self.point_positions.shape[0]} points will be tracked") if self._verbose > 1 else None

        ## Collapse masks into single mask
        print(f"FR: Collapsing mask ROI images into single mask") if self._verbose > 1 else None
        if rois_masks is None:
            self.mask = torch.ones(buffered_video_reader[0][0].shape[:2], dtype=bool)
        else:
            self.mask = torch.as_tensor(np.stack((rois_masks), axis=0).all(axis=0)).type(torch.bool)

        ## Store buffered_video_reader
        self.buffered_video_reader = buffered_video_reader
        ## Prepare video(s)
        self.buffered_video_reader.method_getitem = "by_video"
        if self._contiguous:
            video = self.buffered_video_reader
            video.method_getitem = "continuous"
            video.set_iterator_frame_idx(0)
            self.videos = [video]
        else:
            self.buffered_video_reader.method_getitem = "by_video"
            self.videos = [vid for vid in self.buffered_video_reader]

        ## Initialize mesh distances
        print("FR: Initializing mesh distances") if self._verbose > 1 else None
        p_0 = torch.as_tensor(self.point_positions.copy(), dtype=torch.float32)
        self.neighbors = torch.argsort(torch.linalg.norm(p_0.T[:,:,None] - p_0.T[:,None,:], ord=2, dim=0), dim=1)[:,:self.params_optical_flow["mesh_n_neighbors"]]
        self.d_0 = _vector_distance(torch.as_tensor(self.point_positions.copy(), dtype=torch.float32), self.neighbors)
        
        ## Preallocate points_tracked (will be overwritten with another empty list)
        self.points_tracked = []

        ## Prepare violation tracker
        self._pointIdx_violations_current = np.zeros((self.point_positions.shape[0]), dtype=bool)
        self._violation_event = False
        
        ## Prepare a playback visualizer
        if self._visualize_video:
            print("FR: Preparing playback visualizer") if self._verbose > 1 else None
            self.visualizer = FrameVisualizer(
                display=True,
                handle_cv2Imshow='point_tracking',
                path_save=None,
                frame_height_width=self.buffered_video_reader.frame_height_width,
                error_checking=False,
                verbose=self._verbose,
            )


        ## For FR_Module compatibility
        self.config = {
            "contiguous": self._contiguous,
            "visualize_video": self._visualize_video,
            "params_optical_flow": self.params_optical_flow,
            "params_outlier_handling": self._params_outlier_handling,
            "params_visualization": self._params_visualization,
            "verbose": self._verbose,
        }
        self.run_info = {
        }
        self.run_data = {
            "point_positions": self.point_positions,
            "neighbors": self.neighbors,
            "mesh_d0": self.d_0,
            "mask": self.mask,
        }
        ## Append the self.run_info data to self.run_data
        self.run_data.update(self.run_info)

    # def _make_points(self, roi, point_spacing):
    #     """
    #     Make points within a roi with spacing of point_spacing.
        
    #     Args:
    #         roi (np.ndarray, boolean):
    #             A 2D array of booleans, where True indicates a pixel
    #              that is within the region of interest.
    #         point_spacing (int):
    #             The spacing between points, in pixels.

    #     Returns:
    #         points (np.ndarray, np.float32):
    #             A 2D array of integers, where each row is a point
    #              to track.
    #     """
    #     ## Assert that roi is a 2D array of booleans
    #     assert isinstance(roi, np.ndarray), "FR ERROR: roi must be a numpy array"
    #     assert roi.ndim == 2, "FR ERROR: roi must be a 2D array"
    #     assert roi.dtype == bool, "FR ERROR: roi must be a 2D array of booleans"
    #     ## Warn if point_spacing is not an integer. It will be rounded.
    #     if not isinstance(point_spacing, int):
    #         print("FR WARNING: point_spacing must be an integer. It will be rounded.")
    #         point_spacing = int(round(point_spacing))

    #     ## make point cloud
    #     y, x = np.where(roi)
    #     y_min, y_max = y.min(), y.max()
    #     x_min, x_max = x.min(), x.max()
    #     y_points = np.arange(y_min, y_max, point_spacing)
    #     x_points = np.arange(x_min, x_max, point_spacing)
    #     y_points, x_points = np.meshgrid(y_points, x_points)
    #     y_points = y_points.flatten()
    #     x_points = x_points.flatten()
    #     ## remove points outside of roi
    #     points = np.stack([y_points, x_points], axis=1)
    #     points = points[roi[points[:, 0], points[:, 1]]].astype(np.float32)
    #     ## flip to (x,y)
    #     points = np.fliplr(points)

    #     return points

    def cleanup(self):
        """
        Delete the large data objects that are no longer needed.
        """
        import gc
        print(f"FR: Deleting all attributes")
        while len(self.__dict__.keys()) > 0:
            key = list(self.__dict__.keys())[0]
            del self.__dict__[key]
        gc.collect()
        gc.collect()


    def track_points(self):
        """
        Point tracking workflow.
        Tracks points in videos using specified parameters, roi_points,
         and roi_masks.
        """
        ## Initialize points_tracked
        self.points_tracked = []
        self.violations = []

        ## Set the initial frame_prev as the first frame of the video
        print("FR: Setting initial frame_prev") if self._verbose > 1 else None
        frame_prev = self._format_decordTorchVideo_for_opticalFlow(vid=self.buffered_video_reader.get_frames_from_continuous_index(0), mask=self.mask)[0]
        self.buffered_video_reader.wait_for_loading()
        ## Set the inital points_prev as the original points
        points_prev = self.point_positions.copy()

        
        ## Iterate through videos
        print("FR: Iterating point tracking through videos") if self._verbose > 1 else None
        for ii, video in tqdm(
            enumerate(self.videos), 
            desc='video #', 
            position=1, 
            leave=True, 
            disable=self._verbose < 2, 
            total=len(self.videos)
        ):
            ## If the video is not contiguous, set the iterator to the first frame
            frame_prev = self._format_decordTorchVideo_for_opticalFlow(vid=video.get_frames_from_continuous_index(0), mask=self.mask)[0] if not self._contiguous else frame_prev

            print(f"FR: Iterating through frames of video {ii}") if self._verbose > 2 else None
            points, frame_prev = self._track_points_singleVideo(
                video=video,
                points_prev=points_prev,
                frame_prev=frame_prev,
            )
            self.points_tracked.append(points)
            self.violations.append(self.violations_currentVideo.tocoo())

        print(f"FR: Tracking complete") if self._verbose > 1 else None
        print(f"FR: Placing points_tracked into dictionary self.points_tracked where keys are video indices") if self._verbose > 1 else None
        self.points_tracked = {f"{ii}": points for ii, points in enumerate(self.points_tracked)}
        print(f"FR: Placing violations into dictionary self.violations where keys are video indices") if self._verbose > 1 else None
        self.violations_sparseCOO = {f"{ii}": {'row': v.row, 'col': v.col, 'data': v.data, 'shape': v.shape} for ii, v in enumerate(self.violations)}

        self.violation_fraction = [np.array(v.max(1).toarray()).squeeze().mean() for v in self.violations]

        ## For FR_Module compatibility
        self.run_info["num_videos"] = int(len(self.videos))
        self.run_info["num_frames"] = [int(v.num_frames_total) for v in self.videos]
        self.run_info["num_frames_total"] = int(sum(self.run_info["num_frames"]))
        self.run_info["num_points"] = int(self.num_points)
        self.run_info["violation_fraction"] = [int(v) for v in self.violation_fraction]

        self.run_data["points_tracked"] = self.points_tracked
        self.run_data["violations"] = self.violations_sparseCOO


    def _track_points_singleVideo(
        self,
        video,
        points_prev,
        frame_prev,
    ):
        """
        Track points in a single video.
        
        Args:
            video (BufferedVideoReader):
                A BufferedVideoReader object.
                Should be a single video object, where iterating over
                 the object returns frames. The frames should either
                  be 3D arrays of shape (height, width, channels) or
                  2D arrays of shape (height, width).
                If frames are not np.uint8, they will be converted.
            points_prev (np.ndarray, np.float32):
                A 2D array of np.float32, where each row is a point
                 to track. The points should be in the same order
                 as the points in the previous video. Order (x,y).
            frame_prev (np.ndarray, uint8):
                Previous frame. Should be formatted correctly, as no
                 corrective formatting will be done here.

        Returns:
            points (np.ndarray, np.float32):
                A 2D array of np.float32, where each row is a point
                 that has been tracked. Order (x,y).
            frame_last (np.ndarray, uint8):
                Last frame of the video. Formatted correctly.
        """
        ## Assert that points_prev is a 2D array of integers
        assert isinstance(points_prev, np.ndarray), "FR ERROR: points_prev must be a numpy array"
        assert points_prev.ndim == 2, "FR ERROR: points_prev must be a 2D array"
        points_prev = points_prev.astype(np.float32)

        ## Preallocate points
        points_tracked = np.zeros((len(video), points_prev.shape[0], 2), dtype=np.float32)
        self.violations_currentVideo = scipy.sparse.lil_matrix((len(video), points_prev.shape[0]), dtype=np.bool_)

        self.i_frame = 0
        video.set_iterator_frame_idx(0)
        with tqdm(total=len(video), desc='frame #', position=0, leave=True, disable=self._verbose < 2, mininterval=1.0) as pbar:
            while (self.i_frame < len(video)):
                for frame in video:
                    frame_new = self._format_decordTorchVideo_for_opticalFlow(vid=frame[None,...], mask=self.mask)[0]
                    points_tracked[self.i_frame] = self._track_points_singleFrame(
                        frame_new=frame_new,
                        frame_prev=frame_prev,
                        points_prev=points_prev,
                    )
                    frame_prev = frame_new
                    points_prev = points_tracked[self.i_frame]
                    if self._violation_event:
                        self._violation_event = False
                        self.i_frame = max(self.i_frame - self._params_outlier_handling['framesHalted_before'], 0)
                        frame_prev = self._format_decordTorchVideo_for_opticalFlow(vid=video.get_frames_from_continuous_index(max(self.i_frame-1,0)), mask=self.mask)[0]
                        video.set_iterator_frame_idx(self.i_frame)
                        points_prev = points_tracked[self.i_frame]
                        break

                    pbar.n = self.i_frame
                    self.i_frame += 1
                    ## Update progress bar
                    pbar.update(1)


        ## clear buffered_video_reader
        video.delete_all_slots()

        return points_tracked, frame_prev


    def _track_points_singleFrame(self, frame_new, frame_prev, points_prev):
        """
        Track points in a single frame.
        
        Args:
            frame_new (np.ndarray, uint8):
                A 2D array of integers, where each element is a pixel
                 value. The frame should be the new frame.
            frame_prev (np.ndarray, uint8):
                A 2D array of integers, where each element is a pixel
                 value. The frame should be the previous frame.
            points_prev (np.ndarray, np.float32):
                A 2D array of np.float32, where each row is a point
                 to track. The points should be in the same order
                 as the points in the previous video. Order (x,y).
        """
        ## Call optical flow function
        points_new = self._optical_flow(frame_new=frame_new, frame_prev=frame_prev, points_prev=points_prev)

        ## Update violations
        violations_frame = self._update_violations(points_new=points_new)

        ## Visualize points
        if self._visualize_video:
            self.visualizer.visualize_image_with_points(
                image=cv2.cvtColor(frame_new, cv2.COLOR_GRAY2BGR),
                # points=points_new[None,...].astype(np.int64),
                points=[points_new[self._pointIdx_violations_current].astype(np.int64), points_new[~self._pointIdx_violations_current].astype(np.int64)],
                points_colors=[(0,0,255), (0,255,0)],
                **self._params_visualization,
            )

        return points_new

    def _update_violations(self, points_new):
        """
        Update violations.
        Finds points that violate the max displacement threshold.
        Sets the countdown for these points.
        Updates the countdown for existing violating points.
        Updates the violation event flag.
        Updates the current violation points.

        Args:
            points_new (np.ndarray, np.float32):
                The points that have been tracked. Order (x,y).
                A 2D array of np.float32, where each row is a point
                 to track.
        """
        displacement = points_new - self.point_positions
        pointIdx_violations_new = np.linalg.norm(displacement, axis=1) > self._params_outlier_handling['threshold_displacement']
        ## Find violating neighbors
        pointIdx_violations_new[np.unique(self.neighbors.numpy()[pointIdx_violations_new].reshape(-1))] = True
        ## Determine if a violation event has occurred
        self._violation_event = np.any(pointIdx_violations_new)
        ## Update violation countdowns
        if self._violation_event:
            self.violations_currentVideo[
                max(self.i_frame - self._params_outlier_handling['framesHalted_before'], 0) : \
                min(self.i_frame + self._params_outlier_handling['framesHalted_after'], self.violations_currentVideo.shape[0]),
                pointIdx_violations_new
            ] = True
        ## Find all points that are currently violating
        self._pointIdx_violations_current = self.violations_currentVideo[self.i_frame].toarray().squeeze()


    def _format_decordTorchVideo_for_opticalFlow(self, vid, mask=None):
        """
        Format a decord video array properly for optical flow.
        
        Args:
            vid (decord NDAdarray):
                A 4D array of numbers, where each element is a pixel
                 value. Last dimension should be channels.
                 (batch, height, width, channels)
            mask (np.ndarray, bool):

        Returns:
            vid (np.ndarray, uint8):
                A 3D array of integers, where each element is a pixel
                 value.
                 (batch, height, width)
        """
        ## Collapse channels and mask video
        vid = _helper_format_decordTorchVideo_for_opticalFlow(vid, mask=mask)

        ## Convert to numpy array
        vid = vid.numpy()

        return vid

    def _optical_flow(self, frame_new, frame_prev, points_prev):
        """
        Track points in a single frame using optical flow.
        
        Args:
            frame_new (np.ndarray, uint8):
                A 2D array of integers, where each element is a pixel
                 value. The frame should be the new frame.
            frame_prev (np.ndarray, uint8):
                A 2D array of integers, where each element is a pixel
                 value. The frame should be the previous frame.
            points_prev (np.ndarray, np.float32):
                A 2D array of np.float32, where each row is a point
                 to track. The points should be in the same order
                 as the points in the previous video. Order (x,y).
        """
        ## Call optical flow function
        if self.params_optical_flow['method'] == 'lucas_kanade':
            points_new, status, err = cv2.calcOpticalFlowPyrLK(frame_prev, frame_new, points_prev, None, **self.params_optical_flow['kwargs_method'])
        else:
            raise ValueError("FR ERROR: optical flow method not recognized")

        ## Freeze violating points
        points_new[self._pointIdx_violations_current] = points_prev[self._pointIdx_violations_current]

        ## Apply mesh_rigity force
        points_new -= (_vector_displacement(torch.as_tensor(points_new, dtype=torch.float32), self.d_0, self.neighbors)*self.params_optical_flow['mesh_rigidity']).numpy()

        ## Apply relaxation force
        points_new -= (points_new-self.point_positions)*self.params_optical_flow['relaxation']

        return points_new
    
    def __repr__(self): return f"PointTracker(params_optical_flow={self.params_optical_flow}, visualize_video={self._visualize_video}, verbose={self._verbose})"
    def __getitem__(self, index): return self.points_tracked[index]
    def __len__(self): return len(self.points_tracked)
    def __iter__(self): return iter(self.points_tracked)
    def __next__(self): return next(self.points_tracked)
    

@torch.jit.script
def _helper_format_decordTorchVideo_for_opticalFlow(vid, mask=None):
    ## Collapse channels
    vid = vid.type(torch.float32).mean(dim=-1).type(torch.uint8)

    ## Mask video
    if mask is not None:
        return vid * mask[None, :, :]
    else:
        return vid

@torch.jit.script
def _vector_distance(pi, neighbors):
    """
    Calculate the distance between each point and its neighbors.

    Args:
        pi (torch.Tensor, dtype=torch.float32):
            A 2D array of torch.float32, where each row is a
             (y,x) point.
        neighbors (torch.Tensor, dtype=torch.int64):
            A 2D array of torch.int64 listing the indices of the
             neighbors of each point.

    Returns:
        d (torch.Tensor, dtype=torch.float32):
            A 2D array of torch.float32, where each row is a
             (y,x) vector describing the mean distance between
              each point and its neighbors.
    """
    pm = torch.tile(pi.T[:,:,None], (1,1,neighbors.shape[1]))
    d = pm - pi.T[:, neighbors]
    d2m = d.mean(2).T
    return d2m

@torch.jit.script
def _vector_displacement(di, dj, neighbors):
    """
    Calculate the displacement between each point and its
     neighbors relative to a reference distance (dj).

    Args:
        di (torch.Tensor, dtype=torch.float32):
            A 2D array of torch.float32, where each row is a
             (y,x) point.
        dj (torch.Tensor, dtype=torch.float32):
            A 2D array of torch.float32, where each row is a
             (y,x) distance vector.
        neighbors (torch.Tensor, dtype=torch.int64):
            A 2D array of torch.int64 listing the indices of the
             neighbors of each point.

    Returns:
        d (torch.Tensor, dtype=torch.float32):
            A 2D array of torch.float32, where each row is a
             (y,x) vector describing the mean displacement between
             each point and its neighbors relative to the reference
             distance.
    """
    return _vector_distance(di, neighbors) - dj


"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

import setuptools
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# with open('requirements.txt', 'r') as f:
#     requirements = f.read().splitlines()
# print(requirements)

setuptools.setup(
    name='face_rhythm',
    version='0.1.0',

    description="A pipeline for analysis of facial behavior using optical flow",
    long_description=long_description,
    long_description_content_type='text/markdown',

    # The project's main homepage.
    url='https://github.com/RichieHakim/face-rhythm',

    # Author details
    author='Rich Hakim',

    # Choose your license
    # license='MIT',

    # Supported platforms
    platforms=['Any'],

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',

        # Pick your license as you wish (should match "license" above)
        # 'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # 'Programming Language :: Python :: 2',
        # 'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.6',
        # 'Programming Language :: Python :: 3.7'
    ],

    # What does your project relate to?
    keywords='face beahvior optical flow',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    # packages=setuptools.find_packages(),
    packages=['face_rhythm'],

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    # install_requires=requirements,

    include_package_data=True,
)


from pathlib import Path
import re
import time
from datetime import datetime
import inspect
from typing import Any, Dict, List, Optional, Tuple, Union
import PIL
import warnings

import yaml
import numpy as np

from . import h5_handling
from . import helpers

class FR_Module:
    """
    The superclass for all of the Face Rhythm module classes.
    Allows for saving run_data, run_info, and config files.
    RH 2022
    """
    def __init__(self):
        self.run_info = None
        self.run_data = None

        ## Get module name
        self.module_name = self.__class__.__name__


    def save_config(
        self, 
        path_config=None, 
        overwrite=True, 
        verbose=1
    ):
        """
        Appends the self.config dictionary to the config.yaml file.
        This dictionary is created by the subclass and should contain
         all the parameters used to run the module.
        RH 2022

        Args:
            path_config (str):
                Path to config.yaml file.
            overwrite (bool):
                If True, overwrites fields within the config.yaml file.
            verbose (int):
                Verbosity level. 0 is silent. 1 is print warnings. 2 is print all.
        """
        ## Assert if self.config is not None
        assert self.config is not None, 'FR ERROR: self.config is None. Module likely did not run properly. Please set self.config before saving.'

        ## Assert that path_config is a string, exists, is a file, is a yaml file, and is named properly
        assert isinstance(path_config, str), "FR ERROR: path_config must be a string"
        assert Path(path_config).exists(), "FR ERROR: path_config must exist"
        assert Path(path_config).is_file(), "FR ERROR: path_config must be a file"
        assert Path(path_config).suffix == ".yaml", "FR ERROR: path_config must be a yaml file"
        assert Path(path_config).name == "config.yaml", "FR ERROR: path_config must be named config.yaml"

        config = load_yaml_safe(path_config)
            
        ## Append self.config to module_name key in config.yaml
        if (self.module_name in config.keys()) and not overwrite:
            print(f"FR Warning: Not saving anything. Field exists in dictionary and overwrite==False. '{self.module_name}' is already a field in config.yaml.") if verbose > 0 else None
            return None
        elif (self.module_name in config.keys()) and overwrite:
            print(f"FR Warning: Overwriting field. '{self.module_name}' already in config.yaml.") if verbose > 0 else None
            config[self.module_name] = self.config
        else:
            print(f"FR: Adding '{self.module_name}' to config.yaml") if verbose > 1 else None
            config[self.module_name] = self.config

        ## Update the date_modified field
        config["general"]["date_modified"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        ## Save config.yaml file
        print(f'FR: Saving config.yaml to {path_config}') if verbose > 1 else None
        with open(path_config, 'w') as f:
            yaml.dump(config, f, Dumper=yaml.Dumper, sort_keys=False)


    def save_run_info(
        self, 
        path_run_info=None, 
        path_config=None,
        overwrite=True, 
        verbose=1
    ):
        """
        Appends the self.run_info dictionary to the run_info.yaml file.

        Args:
            path_run_info (str):
                Path to run_info.yaml file.
                Optional. If None, then path_config must be provided, and must
                 contain: config['paths']['project'].
                If the file does not exist, it will be created.
            path_config (str):
                Path to config.yaml file.
                Optional. If None, then path_run_info must be provided.
            overwrite (bool):
                If True, overwrites fields within the run_info.yaml file.
            verbose (int):
                Verbosity level. 0 is silent. 1 is print warnings. 2 is print all.
        """
        ## Assert self.run_info and self.run_data are not None
        assert self.run_info is not None, 'FR ERROR: self.run_info is None. Module likely did not run properly. Please set self.run_info before saving.'
        assert self.run_data is not None, 'FR ERROR: self.run_data is None. Module likely did not run properly. Please set self.run_data before saving.'

        ## Assert that either path_run_info or path_config must be a string, but not both
        assert (path_run_info is not None) and (path_config is None) or (path_run_info is None) and (path_config is not None), "FR ERROR: Either path_run_info or path_config must be specified as a string, but not both"
        ## Get the one that is not None
        path = path_run_info if path_run_info is not None else path_config

        ## Assert that path is a string, exists, is a file, is a json file, and is named properly
        assert isinstance(path, str), "FR ERROR: path_run_info must be a string"
        assert Path(path).exists(), "FR ERROR: path_run_info must exist"
        assert Path(path).is_file(), "FR ERROR: path_run_info must be a file"
        if path_run_info is not None:
            assert Path(path_run_info).name == "run_info.json", "FR ERROR: path_run_info must be named run_info.json"
        if path_config is not None:
            assert Path(path_config).name == "config.yaml", "FR ERROR: path_config must be named config.yaml"

        ## Set path_run_info. Get from config if path_run_info is None
        path_run_info = load_yaml_safe(path_config)["paths"]["run_info"] if path_run_info is None else path_run_info

        ## Check if file exists and load it if it does
        ## If directory to file does not exist, create it
        if Path(path_run_info).exists()==False:
            print(f'FR: No existing run_info.json file found in {path_run_info}. \n Creating new run_info.json at {path_run_info}') if verbose > 0 else None
            Path(path_run_info).parent.mkdir(parents=True, exist_ok=True)
            run_info = {}
        else:
            print(f'FR: Loading file {path_run_info}') if verbose > 1 else None
            run_info = helpers.json_load(path_run_info, mode='r')
            
        ## Append self.run_info to module_name key in run_info.json
        if (self.module_name in run_info.keys()) and not overwrite:
            print(f"FR Warning: Not saving anything. Field exists in dictionary and overwrite==False. '{self.module_name}' is already a field in run_info.json.") if verbose > 0 else None
        elif (self.module_name in run_info.keys()) and overwrite:
            print(f"FR Warning: Overwriting field. '{self.module_name}' is already a field in the run_info.json dictionary.") if verbose > 0 else None
            run_info[self.module_name] = self.run_info
        else:
            print(f"FR: Adding '{self.module_name}' field to run_info.json") if verbose > 1 else None
            run_info[self.module_name] = self.run_info
        
        ## Save run_info.json file
        print(f'FR: Saving run_info.json to {path_run_info}') if verbose > 1 else None
        helpers.json_save(run_info, path_run_info, mode='w')
            

    def save_run_data(
        self, 
        path_run_data=None, 
        path_config=None,
        overwrite=True, 
        use_compression=False,
        track_order=True,
        verbose=1
    ):
        """
        Appends the self.run_data dictionary to a .h5 file in the
         .../project/analaysis_files/'object method name'.h5.
        The self.run_data dictionary is created by the subclass and should contain
         all the data generated by the module.
        The project directory should already exist and can be created using
         the face_rhythm.project.prepare_project function.
        RH 2022

        Args:
            path_run_data (str):
                Path to .h5 file.
                Optional. If None, then path_config must be provided, and must
                 contain: config['paths']['project']
                If the file does not exist, it will be created.
            path_config (str):
                Path to config.yaml file.
                Optional. If None, then path_run_data must be provided.
                Should contain: config['paths']['project']. path_run_data will be:
                 .../config['paths']['project']/analysis_files/'object method name'.h5
            overwrite (bool):
                If True, overwrites fields within the .h5 file.
            use_compression (bool):
                If True, uses compression when saving the .h5 file.
            track_order (bool):
                If True, tracks the order of the data in the .h5 file.
            verbose (int):
                Verbosity level. 
                0: silent
                1: print warnings
                2: print all info
        """
        ## Assert self.run_data is not None
        assert self.run_data is not None, 'FR ERROR: self.run_data is None. Module likely did not run properly. Please set self.run_data before saving.'

        ## Assert that either path_run_data or path_config must be a string, but not both
        assert (path_run_data is not None) and (path_config is None) or (path_run_data is None) and (path_config is not None), "FR ERROR: Either path_run_info or path_config must be specified as a string, but not both"
        ## If path_run_data is None, then path_config must be a string, exist, be a file, be a yaml file, and contain the project directory
        if path_run_data is None:
            assert isinstance(path_config, str), "FR ERROR: path_config must be a string"
            assert Path(path_config).exists(), "FR ERROR: path_config must exist"
            assert Path(path_config).is_file(), "FR ERROR: path_config must be a file"
            assert Path(path_config).suffix == ".yaml", "FR ERROR: path_config must be a yaml file"
            config = load_yaml_safe(path_config)
            assert 'project' in config['paths'].keys(), "FR ERROR: config['paths']['project'] must exist in path_config"
            path_run_data = str(Path(config['paths']['project']) / 'analysis_files' / f'{self.module_name}.h5')
            print(f"FR: Using project directory (config['paths']['project']) from config.yaml to make run_data path: {path_run_data}") if verbose > 1 else None

        ## Assert path_run_data is a string
        assert isinstance(path_run_data, str), "FR ERROR: path_run_data must be a string"
        if path_run_data is not None:
            print(f"FR WARNING: path_run_data file is expected to be named '{self.module_name+'.h5'}' if it is part of a project. Please make sure this is correct.") if verbose > 0 else None
        ## If a file exists and overwrite is False, then print a warning and cancel out
        ## If a file exists and overwrite is True, then print a warning and continue
        if Path(path_run_data).exists():
            if not overwrite:
                print(f'FR Warning: Not saving anything. File exists and overwrite==False. {path_run_data} already exists.') if verbose > 0 else None
                return None
            else:
                print(f'FR Warning: Overwriting file. File: {path_run_data} already exists.') if verbose > 0 else None


        ## Create directory if it does not exist
        if not Path(path_run_data).parent.exists():
            print(f'FR: Creating directory {Path(path_run_data).parent}') if verbose > 1 else None
            Path(path_run_data).parent.mkdir(parents=True)

        ## Try to save run_data to .h5 file. If we get an error that it failed because the ile is already open, then search for all open h5py.File objects and close them.
        print(f'FR: Saving run_data to {path_run_data}') if verbose > 1 else None
        try:
            h5_handling.simple_save(
                dict_to_save=self.run_data, 
                path=path_run_data, 
                use_compression=use_compression, 
                track_order=track_order,
                write_mode=('w' if overwrite else 'w-'), 
                verbose=verbose>1
            )
        except OSError as e:
            if re.search('Unable.*already open', str(e)):
                print(f'FR Warning: {path_run_data} is already open. Closing all open h5py.File objects and trying again.') if verbose > 0 else None
                h5_handling.close_all_h5()
                h5_handling.simple_save(
                    dict_to_save=self.run_data, 
                    path=path_run_data, 
                    use_compression=use_compression, 
                    track_order=track_order,
                    write_mode=('w' if overwrite else 'w-'), 
                    verbose=verbose>1
                )
            else:
                raise e
        
        ## Assert that the file exists
        assert Path(path_run_data).exists(), "FR ERROR: path_run_data must exist"
        ## Warn if it was not saved recently
        if (time.time() - Path(path_run_data).stat().st_mtime) > 1:
            print(f'FR Warning: Saving run_data may have failed. {path_run_data} was not saved recently.') if verbose > 0 else None

        

def load_yaml_safe(path, verbose=0):
    """
    loads yaml file

    Args:
        path (str): 
            path to .yaml file

    Returns:
        (dict): 
            yaml file as a dictionary

    """
    print(f'FR: Loading file {path}') if verbose > 1 else None
    try:
        with open(path, 'r') as f:
            return yaml.load(f, Loader=yaml.FullLoader)
    except:
        print(f'FR Warning: Failed to load {path} with Loader=yaml.FullLoader. A field is likely not yaml compatible. Trying with yaml.Loader.')
        with open(path, 'r') as f:
            return yaml.load(f, Loader=yaml.Loader)

def load_config_file(path, verbose=0):
    """
    Loads config.yaml file

    Args:
        path (str): 
            path to config.yaml file

    Returns:
        (dict): 
            config.yaml file as a dictionary

    """
    return load_yaml_safe(path, verbose=verbose)
def load_run_info_file(path, verbose=0):
    """
    Loads run_info.json file

    Args:
        path (str): 
            path to run_info.json file

    Returns:
        (dict): 
            run_info.json file as a dictionary

    """
    return helpers.json_load(path, mode='r')


class Saver_Viz_Base:
    """
    Super class for saving visualizations i.e. (Figuer_Saver, Image_Saver)
    """
    def __init__(
        self,
        path_config: str=None,
        dir_save: str=None,
        formats_save: list=['png'],
        kwargs_method: dict={},
        overwrite: bool=False,
        verbose: int=1,
    ):
        """
        Initializes Visualization_Saver object.

        Args:
            path_config (str):
                Path to config.yaml file
                Optional if dir_save is specified
            dir_save (str):
                Directory to save visualizations
                Optional if path_config is specified
            formats_save (list):
                List of formats to save visualizations as.
                Depedenent on the method used to save the visualization.
            kwargs_method (dict):
                Dictionary of keyword arguments to pass to the method 
                 used to save the visualization.
            overwrite (bool):
                Whether to overwrite existing files.
            verbose (int):
                Level of verbosity. 0 prints nothing. 1 prints warnings.
                 2 prints warnings and information.
        """
        ## Validate inputs
        assert isinstance(path_config, str) or isinstance(dir_save, str), "FR ERROR: Either path_config or dir_save must be specified as a string."
        if path_config is not None:
            assert Path(path_config).exists(), "FR ERROR: path_config must exist"
        if isinstance(formats_save, str):
            formats_save = [formats_save]
        assert isinstance(formats_save, list), "FR ERROR: formats_save must be a list"
        assert all([isinstance(f, str) for f in formats_save]), "FR ERROR: formats_save must be a list of strings"

        ## Set attributes
        self.path_config = path_config
        self.dir_save = dir_save
        self.formats_save = formats_save
        self.kwargs_method = kwargs_method
        self.overwrite = overwrite
        self.verbose = verbose

        ## Load config file
        self.dir_save = str(Path(load_config_file(self.path_config)['paths']['project']) / 'visualizations') if dir_save is None else dir_save

        ## Create directory if it does not exist
        if not Path(self.dir_save).exists():
            Path(self.dir_save).mkdir(parents=True, exist_ok=True)
            print(f'FR: Created directory {self.dir_save}') if verbose > 0 else None

    def _save_single(
        self,
        name_save: str,
        obj_save: object,
        fn_save: callable,
        kwargs_method: dict={},
        format_save: str=None,
    ):
        """
        Saves a single visualization.

        Args:
            name_save (str):
                Name of the file to save the visualization as.
            obj_save (object):
                Object to save.
            fn_save (callable):
                Function to use to save the visualization.
                Should take args: (obj_save, path_save, format_save, **kwargs_method)
            kwargs_method (dict):
                Dictionary of keyword arguments to pass to the method 
                 used to save the visualization.
            format_save (str):
                Format to save the visualization as.
                If None, then the default format is used.
        """
        ## Validate inputs
        assert isinstance(name_save, str), "FR ERROR: name_save must be a string"
        assert isinstance(obj_save, object), "FR ERROR: obj_save must be an object"
        assert callable(fn_save), "FR ERROR: fn_save must be callable"
        assert isinstance(kwargs_method, dict), "FR ERROR: kwargs_method must be a dictionary"
        assert isinstance(format_save, str), "FR ERROR: format_save must be a string"

        ## Set kwargs_method
        kwargs_method = {**self.kwargs_method, **kwargs_method}

        ## Set and prepare path to save
        path_save = str(Path(self.dir_save).resolve() / f"{name_save}.{format_save}")
        helpers.prepare_filepath_for_saving(path_save, mkdir=True, allow_overwrite=self.overwrite)

        ## assert that fn_save has the correct kwargs
        args_fn_save = inspect.getfullargspec(fn_save).args
        assert all([k in args_fn_save for k in ['obj_save', 'path_save', 'format_save']]), "FR ERROR: fn_save must have args: ['obj_save', 'path_save', 'format_save']"

        ## Save visualization
        fn_save(
            obj_save=obj_save,
            path_save=path_save,
            format_save=format_save,
            kwargs_method=kwargs_method,
        )

    def _inherit_from_attrs(self, vars, attrs):
        for var, attr in zip(vars, attrs):
            if var is None:
                assert hasattr(self, attr), f"FR ERROR: {attr} must be specified in either the constructor or the method call"
                var = getattr(self, attr)
            yield var

    def __repr__(self):
        return f"Figure_Saver(path_config={self.path_config}, dir_save={self.dir_save}, formats_save={self.formats_save}, kwargs_method={self.kwargs_method}, overwrite={self.overwrite}, verbose={self.verbose})"


class Figure_Saver(Saver_Viz_Base):
    """
    Class for saving figures
    RH 2022
    """
    def __init__(
        self,
        path_config: str=None,
        dir_save: str=None,
        formats_save: list=['png'],
        kwargs_savefig: dict={
            'bbox_inches': 'tight',
            'pad_inches': 0.1,
            'transparent': True,
            'dpi': 300,
        },
        overwrite: bool=False,
        verbose: int=1,
    ):
        """
        Initializes Figure_Saver object

        Args:
            path_config (str):
                Path to config.yaml file. If None, then path_save must
                be specified.
            dir_save (str):
                Directory to save the figure. Used if path_config is None.
                Must be specified if path_config is None.
            formats_save (list of str):
                Format(s) to save the figure. Default is 'png'.
                Others: ['png', 'svg', 'eps', 'pdf']
            overwrite (bool):
                If True, then overwrite the file if it exists.
            kwargs_savefig (dict):
                Keyword arguments to pass to fig.savefig().
            verbose (int):
                Verbosity level.
                0: No output.
                1: Warning.
                2: All info.
        """
        ## Initialize super
        super().__init__(
            path_config=path_config,
            dir_save=dir_save,
            formats_save=formats_save,
            overwrite=overwrite,
            verbose=verbose,
        )

        ## Set kwargs_savefig
        self.kwargs_savefig = kwargs_savefig

        self.__call__ = self.save_figure

    def save_figure(
        self,
        fig,
        name_save: str=None,
        dir_save: str=None,
        formats_save: str=None,
        kwargs_savefig: dict=None,
    ):
        """
        Saves a single figure.

        Args:
            fig (matplotlib.figure.Figure):
                Figure to save.
            name_save (str):
                Name of the file to save the figure as. 
                If None, then the name of the figure is used.
            dir_save (str):
                Directory to save the figure. If None, then the directory
                 specified in the initialization is used.
            formats_save (str):
                Formats to save the figure as. If None, then the format
                 specified in the initialization is used.
            kwargs_savefig (dict):
                Keyword arguments to pass to fig.savefig().
        """
        import matplotlib

        ## Set missing inputs
        name_save = name_save if name_save is not None else fig.get_label()
        name_save = 'fig' if len(name_save) == 0 else name_save        
        dir_save, formats_save, kwargs_savefig = self._inherit_from_attrs(
            vars=[dir_save, formats_save, kwargs_savefig], 
            attrs=['dir_save', 'formats_save', 'kwargs_savefig'],
        )
        formats_save = [formats_save] if not isinstance(formats_save, list) else formats_save

        ## Validate inputs
        assert isinstance(fig, matplotlib.figure.Figure), "FR ERROR: fig must be a matplotlib.figure.Figure"

        ## Save figure
        fn_save = lambda obj_save, path_save, format_save, kwargs_method: obj_save.savefig(path_save, format=format_save, **kwargs_method)
        for format_save in formats_save:
            self._save_single(
                name_save=name_save,
                obj_save=fig,
                fn_save=fn_save,
                kwargs_method=kwargs_savefig,
                format_save=format_save,
            )

class Image_Saver(Saver_Viz_Base):
    def __init__(
        self,
        path_config: str=None,
        dir_save: str=None,
        formats_save: list=['png'],
        kwargs_PIL_save: dict={
        },
        overwrite: bool=False,
        verbose: int=1,
    ):
        """
        Initializes Image_Saver object

        Args:
            path_config (str):
                Path to config.yaml file. If None, then path_save must
                be specified.
            dir_save (str):
                Directory to save the figure. Used if path_config is None.
                Must be specified if path_config is None.
            formats_save (list of str):
                Format(s) to save the figure. Default is 'png'.
                Others: ['png', 'svg', 'eps', 'pdf']
            kwargs_PIL_save (dict):
                Keyword arguments to pass to PIL.Image.save().
            overwrite (bool):
                If True, then overwrite the file if it exists.
            verbose (int):
                Verbosity level.
                0: No output.
                1: Warning.
                2: All info.
        """
        ## Initialize super
        super().__init__(
            path_config=path_config,
            dir_save=dir_save,
            formats_save=formats_save,
            overwrite=overwrite,
            verbose=verbose,
        )

        ## Set kwargs_PIL_save
        self.kwargs_PIL_save = kwargs_PIL_save

        self.__call__ = self.save_image

    def save_image(
        self,
        array_image,
        name_save: str=None,
        dir_save: str=None,
        formats_save: str=None,
        kwargs_PIL_save: dict=None,
    ):
        """
        Saves a single image.

        Args:
            array_image (numpy.ndarray):
                Image to save. If float, then should be between 0 and 1. 
                Will be * 255 and cast to uint8.
                If int, then should be between 0 and 255. Will be cast to uint8.
            name_save (str):
                Name of the file to save the figure as. 
                If None, then the name of the figure is used.
            dir_save (str):
                Directory to save the figure. If None, then the directory
                 specified in the initialization is used.
            formats_save (str):
                Formats to save the figure as. If None, then the format
                 specified in the initialization is used.
            kwargs_PIL_save (dict):
                Keyword arguments to pass to PIL.Image.save(). 
        """

        ## Set missing inputs
        name_save = name_save if name_save is not None else 'image'
        dir_save, formats_save, kwargs_PIL_save = self._inherit_from_attrs(
            vars=[dir_save, formats_save, kwargs_PIL_save],
            attrs=['dir_save', 'formats_save', 'kwargs_PIL_save'],
        )
        formats_save = [formats_save] if not isinstance(formats_save, list) else formats_save
        
        ## Validate inputs
        array_image = self._prepare_array_image(array_image)

        ## Save image
        for format_save in formats_save:
            self._save_single(
                name_save=name_save,
                obj_save=array_image,
                fn_save=self._fn_save_single_image,
                kwargs_method=kwargs_PIL_save,
                format_save=format_save,
            )
        
    def save_gif(
        self,
        array_images,
        name_save: str=None,
        dir_save: str=None,
        frame_rate: float=5.0,
        loop: int=True,
        optimize: bool=True,
        kwargs_PIL_save: dict=None,
    ):
        """
        Saves multiple images as a gif using PIL.
        """
        ## Set missing inputs
        name_save = name_save if name_save is not None else 'image'
        dir_save, kwargs_PIL_save = self._inherit_from_attrs(
            vars=[dir_save, kwargs_PIL_save],
            attrs=['dir_save', 'kwargs_PIL_save'],
        )
        formats_save = ['gif']
        
        ## Validate inputs
        assert isinstance(array_images, list), "FR ERROR: array_images must be a list"
        
        kwargs_PIL_save['optimize'] = optimize

        kwargs_method = {
            'frame_rate': frame_rate,
            'loop': loop,
            'kwargs_PIL_save': kwargs_PIL_save,
        }

        ## Save gif
        for format_save in formats_save:
            self._save_single(
                name_save=name_save,
                obj_save=array_images,
                fn_save=self._fn_save_gif,
                kwargs_method=kwargs_method,
                format_save=format_save,
            )

    def _fn_save_single_image(self, obj_save, path_save, format_save, kwargs_method):
        """
        Converts a single 3D numpy.ndarray with shape[-1] == 3 or 1 to a PIL.Image
         and saves it.
        """
        format_LUT = {
            'jpg': 'JPEG',
            'tif': 'TIFF',
        }
        format_save = format_LUT.get(format_save, format_save)
        obj_save = PIL.Image.fromarray(obj_save, mode='RGB') if obj_save.shape[-1] == 3 else PIL.Image.fromarray(obj_save, mode='L')
        obj_save.save(path_save, format=format_save, **kwargs_method)

    def _fn_save_gif(self, obj_save, path_save, format_save, kwargs_method):
        """
        Converts a list of 3D numpy.ndarrays with shape[-1] == 3 or 1 to a PIL.Image
         and saves it.
        """
        helpers.save_gif(
            array=obj_save, 
            path=path_save, 
            frameRate=kwargs_method['frame_rate'],
            loop=kwargs_method['loop'],
            backend='PIL',
            kwargs_backend=kwargs_method['kwargs_PIL_save'],
        )

        
    
    def _prepare_array_image(self, array_image):
        """
        Converts an input array_image from a 2D or 3D numpy.ndarray of floats
         between 0 and 1 or ints between 0 and 255 to a 3D numpy.ndarray of
         uint8s between 0 and 255.
        """
        ## Validate inputs
        assert isinstance(array_image, np.ndarray), "FR ERROR: array_image must be a numpy.ndarray"
        assert array_image.ndim in [2, 3], "FR ERROR: array_image must be a 2D or 3D numpy.ndarray"

        ## Prepare array_image
        if array_image.ndim == 2:
            array_image = np.expand_dims(array_image, axis=-1)

        if np.issubdtype(array_image.dtype, np.floating):
            assert np.all((0 <= array_image) & (array_image <= 1)), "FR ERROR: images must be between 0 and 1"
            array_image = (array_image * 255).astype(np.uint8)
        elif np.issubdtype(array_image.dtype, np.integer):
            assert np.all((0 <= array_image) & (array_image <= 255)), "FR ERROR: images must be between 0 and 255"
            array_image = array_image.astype(np.uint8)
        else:
            raise ValueError("FR ERROR: array_image.dtype must be float or int")

        return array_image

    

def system_info(verbose: bool = False,) -> Dict:
    """
    Checks and prints the versions of various important software packages.
    RH 2022

    Args:
        verbose (bool): 
            Whether to print the software versions. 
            (Default is ``False``)

    Returns:
        (Dict): 
            versions (Dict):
                Dictionary containing the versions of various software packages.
    """
    ## Operating system and version
    import platform
    def try_fns(fn):
        try:
            return fn()
        except:
            return None
    fns = {key: val for key, val in platform.__dict__.items() if (callable(val) and key[0] != '_')}
    operating_system = {key: try_fns(val) for key, val in fns.items() if (callable(val) and key[0] != '_')}
    print(f'== Operating System ==: {operating_system["uname"]}') if verbose else None

    ## CPU info
    try:
        import cpuinfo
        import multiprocessing as mp
        # cpu_info = cpuinfo.get_cpu_info()
        cpu_n_cores = mp.cpu_count()
        cpu_brand = cpuinfo.cpuinfo.CPUID().get_processor_brand(cpuinfo.cpuinfo.CPUID().get_max_extension_support())
        cpu_info = {'n_cores': cpu_n_cores, 'brand': cpu_brand}
        if 'flags' in cpu_info:
            cpu_info['flags'] = 'omitted'
    except Exception as e:
        warnings.warn(f'RH WARNING: unable to get cpu info. Got error: {e}')
        cpu_info = 'Error: Failed to get'
    print(f'== CPU Info ==: {cpu_info}') if verbose else None

    ## RAM
    import psutil
    ram = psutil.virtual_memory()
    print(f'== RAM ==: {ram}') if verbose else None

    ## User
    import getpass
    user = getpass.getuser()

    ## GPU
    try:
        import GPUtil
        gpus = GPUtil.getGPUs()
        gpu_info = {gpu.id: gpu.__dict__ for gpu in gpus}
    except Exception as e:
        warnings.warn(f'RH WARNING: unable to get gpu info. Got error: {e}')
        gpu_info = 'Error: Failed to get'
    print(f'== GPU Info ==: {gpu_info}') if verbose else None
    
    ## Conda Environment
    import os
    if 'CONDA_DEFAULT_ENV' not in os.environ:
        conda_env = 'None'
    else:
        conda_env = os.environ['CONDA_DEFAULT_ENV']
    print(f'== Conda Environment ==: {conda_env}') if verbose else None

    ## Python
    import sys
    python_version = sys.version.split(' ')[0]
    print(f'== Python Version ==: {python_version}') if verbose else None

    ## GCC
    import subprocess
    try:
        gcc_version = subprocess.check_output(['gcc', '--version']).decode('utf-8').split('\n')[0].split(' ')[-1]
    except Exception as e:
        warnings.warn(f'RH WARNING: unable to get gcc version. Got error: {e}')
        gcc_version = 'Faled to get'
    print(f'== GCC Version ==: {gcc_version}') if verbose else None
    
    ## PyTorch
    import torch
    torch_version = str(torch.__version__)
    print(f'== PyTorch Version ==: {torch_version}') if verbose else None
    ## CUDA
    if torch.cuda.is_available():
        cuda_version = torch.version.cuda
        cudnn_version = torch.backends.cudnn.version()
        torch_devices = [f'device {i}: Name={torch.cuda.get_device_name(i)}, Memory={torch.cuda.get_device_properties(i).total_memory / 1e9} GB' for i in range(torch.cuda.device_count())]
        print(f"== CUDA Version ==: {cuda_version}, CUDNN Version: {cudnn_version}, Number of Devices: {torch.cuda.device_count()}, Devices: {torch_devices}, ") if verbose else None
    else:
        cuda_version = None
        cudnn_version = None
        torch_devices = None
        print('== CUDA is not available ==') if verbose else None

    ## all packages in environment
    import pkg_resources
    pkgs_dict = {i.key: i.version for i in pkg_resources.working_set}

    ## face_rhythm
    import face_rhythm
    import time
    face_rhythm_version = face_rhythm.__version__
    face_rhythm_fileDate = time.ctime(os.path.getctime(pkg_resources.get_distribution("face_rhythm").location))
    face_rhythm_stuff = {'version': face_rhythm_version, 'date_installed': face_rhythm_fileDate}
    print(f'== face_rhythm Version ==: {face_rhythm}') if verbose else None
    print(f'== face_rhythm date installed ==: {face_rhythm_fileDate}') if verbose else None

    ## get datetime
    from datetime import datetime
    dt = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    versions = {
        'datetime': dt,
        'face_rhythm': face_rhythm_stuff,
        'operating_system': operating_system,
        'cpu_info': cpu_info,  ## This is the slow one.
        'user': user,
        'ram': ram,
        'gpu_info': gpu_info,
        'conda_env': conda_env,
        'python': python_version,
        'gcc': gcc_version,
        'torch': torch_version,
        'cuda': cuda_version,
        'cudnn': cudnn_version,
        'torch_devices': torch_devices,
        'pkgs': pkgs_dict,
    }

    return versions

def batch_run(paths_scripts, 
                params_list, 
                sbatch_config_list, 
                max_n_jobs=2,
                dir_save='/n/data1/hms/neurobio/sabatini/rich/analysis/', 
                name_save='jobNum_', 
                verbose=True,
                ):
    r"""
    FROM BNPM
    Run a batch of jobs.
    Workflow 1: run a single script over a sweep of parameters
        - Make a script that takes in the set of parameters
           you wish to sweep over as variables.
        - Prepend the script to take in string arguments
           pointing to a param_config file (maybe a dict).
           See paths_scripts Arg below for details.
        - Save the script in .py file.
        - In a new script, call this function (batch_run)
        - A new job will be run for each item in params_list
            - Each job will make a new directory, and within
               it will save (1) a .json file containing the 
               parameters used, and (2) the .sh file that 
               was run.
        - Save output files using the 'dir_save' argument

    Alternative workflows where you have multiple different
     scripts or different config files are also possible.

    RH 2021

    Args:
        paths_scripts (List):
            - List of script paths to run.
            - List can contain either 1 or n_jobs items.
            - Each script must save its results it's own way
               using a relative path (see 'dir_save' below)
            - Each script should contain the following to handle
               input arguments specified by the user and this
               function, DEMO:
                ```
                import sys
                    path_script, path_params, dir_save = sys.argv
                
                import json
                with open(path_params, 'r') as f:
                    params = json.load(f)
                ```                
            - It's also good practice to save the script .py file
               within dir_save DEMO:
                ```
                import shutil
                shutil.copy2(
                    path_script, 
                    str(Path(dir_save) / Path(path_script).name)
                    );
                ```
        params_list (List):
            - Parameters (arguments) to be used
            - List can contain either 1 or n_jobs items.
            - Each will be saved as a .json file (so nothing too big)   
            - Will be save into each inner/job directory and the path
               will be passed to the script for each job.
        sbatch_config_list (List):
            - List of string blocks containing the arguments to 
               pass for each job/script.
            - List can contain either 1 or n_jobs items.
            - Must contain: python "$@" at the bottom (to take in 
               arguments), and raw string must have '\n' to signify
               line breaks.
               Demo: '#!/usr/bin/bash
                    #SBATCH --job-name=python_01
                    #SBATCH --output=jupyter_logs/python_01_%j.log
                    #SBATCH --partition=priority
                    #SBATCH -c 1
                    #SBATCH -n 1
                    #SBATCH --mem=1GB
                    #SBATCH --time=0-00:00:10

                    unset XDG_RUNTIME_DIR

                    cd /n/data1/hms/neurobio/sabatini/rich/

                    date

                    echo "loading modules"
                    module load gcc/9.2.0 cuda/11.2

                    echo "activating environment"
                    source activate ROI_env

                    echo "starting job"
                    python "$@" '
        max_n_jobs (int):
            - Maximum number of jobs that can be called
            - Used as a safety precaution
            - Be careful that params_list has the right len
        dir_save (str or Path):
            - Outer directory to save results to.
            - Will be created if it does not exist.
            - Will be populated by folders for each job
            - Will be sent to the script for each job as the
               third argument. See paths_scripts demo for details.
        name_save (str or List):
            - Name of each job (used as inner directory name)
            - If str, then will be used for all jobs 
            - Job iteration always appended to the end.
            - If List, then must have len(params_list) items.
        verbose (bool):
            - Whether or not to print progress
    """
    import json
    import os
    import shutil

    # make sure the arguments are matched in length
    n_jobs = max(len(paths_scripts), len(params_list), len(sbatch_config_list))
    if max_n_jobs is not None:
        if n_jobs > max_n_jobs:
            raise ValueError(f'Too many jobs requested: max_n_jobs={n_jobs} > n_jobs={max_n_jobs}')

    def rep_inputs(item, n_jobs):
        if len(item)==1 and (n_jobs>1):
            return helpers.Lazy_repeat_item(item[0], pseudo_length=n_jobs)
        else:
            return item

    paths_scripts      = rep_inputs(paths_scripts,   n_jobs)
    params_list        = rep_inputs(params_list,  n_jobs)
    sbatch_config_list = rep_inputs(sbatch_config_list, n_jobs)
    name_save          = rep_inputs([name_save], n_jobs)

    # setup the save path
    Path(dir_save).mkdir(parents=True, exist_ok=True)
    dir_save = Path(dir_save).resolve()

    # run the jobs
    for ii in range(n_jobs):
        dir_save_job = dir_save / f'{name_save[ii]}{ii}'
        dir_save_job.mkdir(parents=True, exist_ok=True)
        # save the shell scripts
        save_path_sbatchConfig = dir_save_job / 'sbatch_config.sh'
        with open(save_path_sbatchConfig, 'w') as f:
            f.write(sbatch_config_list[ii])
        # save the script
        path_script_job = dir_save_job / Path(paths_scripts[ii]).name
        shutil.copy2(src=paths_scripts[ii], dst=str(path_script_job));
        # save the parameters        
        path_params_job = dir_save_job / 'params.json'
        with open(path_params_job, 'w') as f:
            json.dump(params_list[ii], f)
    
        # run the job
        if verbose:
            print(f'Submitting job: {name_save[ii]} {ii}')
        # ! sbatch --job-name=${name_save}_${ii} --output=${dir_save_job}/log.txt --error=${dir_save_job}/err.txt --time=${sbatch_config_list[ii]["time"]} --mem=${sbatch_config_list[ii]["mem"]} --cpus-per-task=${sbatch_config_list[ii]["cpus"]} --wrap="${paths_scripts[ii]} ${params_list[ii]} ${sbatch_config_list[ii]} ${dir_save_job}"
        os.system(f'sbatch {save_path_sbatchConfig} {path_script_job} --path_params {path_params_job} --directory_save {dir_save_job}')
